from typing import Any, ClassVar, Optional, Type

from dictdaora import DictDaora
from jsondaora import as_typed_dict_field


BUILTIN_TYPE = type


class _Header(DictDaora):  # type: ignore
    def __init__(self, value: Any):
        value = as_typed_dict_field(value, 'value', type(self).type)
        super().__init__(value=value)


def header(
    type: Type[Any], http_name: Optional[str] = None
) -> Type[DictDaora]:
    annotations = {
        'type': ClassVar[Type[Any]],
        'http_name': ClassVar[Optional[str]],
        'value': Any,
    }
    return BUILTIN_TYPE(
        'Header',
        (_Header,),
        {'__annotations__': annotations, 'http_name': http_name, 'type': type},
    )



import io
from typing import IO, Optional, Type

from dictdaora import DictDaora
from jsondaora import jsondaora


try:
    import gzip
except Exception:
    gzip = None  # type: ignore


class _GZipFactory(DictDaora):  # type: ignore
    mode: str
    compresslevel: int
    encoding: Optional[str]
    errors: Optional[str]
    newline: Optional[str]
    value: bytes

    def open(self) -> IO[bytes]:
        if self.value:
            return gzip.open(
                io.BytesIO(self.value),
                self.mode,
                self.compresslevel,
                self.encoding,
                self.errors,
                self.newline,
            )

        raise ValueError(self.value)


def gzip_body(
    mode: str = 'rb',
    compresslevel: int = 9,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
) -> Type[_GZipFactory]:
    cls_attributes = dict(
        mode=mode,
        compresslevel=compresslevel,
        encoding=encoding,
        errors=errors,
        newline=newline,
    )
    factory = type('GZipFactory', (_GZipFactory,), cls_attributes)
    return jsondaora(factory)  # type: ignore



from typing import TypedDict

from jsondaora import integer, jsondaora, string

from apidaora import appdaora, header, route


Age = header(type=int)


@jsondaora
class You(TypedDict):
    name: str
    last_name: str
    location: string(max_length=100)
    age: integer(minimum=18)


@jsondaora
class ReqBody(TypedDict):
    last_name: str


@jsondaora
class HelloOutput(TypedDict):
    hello_message: str
    abot_you: You


@route.put('/hello/{name}')
async def hello_controller(
    name: str, location: str, age: Age, body: ReqBody
) -> HelloOutput:
    you = You(name=name, location=location, age=age.value, **body)
    return HelloOutput(
        hello_message=await hello_message(name, location), about_you=you
    )


async def hello_message(name: str, location: str) -> str:
    return f'Hello {name}! Welcome to {location}!'


app = appdaora(hello_controller)



"""
ASGI App using dataclasses module for request/response objects
"""

__version__ = '0.10.0'

from apidaora.app import appdaora
from apidaora.bodies import gzip_body
from apidaora.content import ContentType
from apidaora.exceptions import BadRequestError
from apidaora.header import header
from apidaora.method import MethodType
from apidaora.responses import html, json, text
from apidaora.route.decorator import route


__all__ = [
    BadRequestError.__name__,
    ContentType.__name__,
    MethodType.__name__,
    appdaora.__name__,
    gzip_body.__name__,
    header.__name__,
    html.__name__,
    json.__name__,
    text.__name__,
    route.__name__,
]



import functools
from typing import Any, Callable, Union

from ..asgi.router import Controller
from ..controllers.background_task import BackgroundTask, make_background_task
from ..exceptions import InvalidRouteArgumentsError, MethodNotFoundError
from ..method import MethodType
from .factory import make_route


class _RouteDecorator:
    def __getattr__(self, attr_name: str) -> Any:
        if attr_name == '__name__':
            return type(self).__name__

        if attr_name == 'background':
            brackground = True

        else:
            brackground = False
            method = attr_name.upper()

            if method not in MethodType.__members__:
                raise MethodNotFoundError(attr_name)

        def decorator(
            path_pattern: str, **kwargs: Any
        ) -> Callable[[Callable[..., Any]], Controller]:
            if len(kwargs) > 0 and tuple(kwargs.keys()) != (
                'tasks_repository',
            ):
                raise InvalidRouteArgumentsError(kwargs)

            @functools.wraps(make_route)
            def wrapper(
                controller: Callable[..., Any]
            ) -> Union[Controller, BackgroundTask]:
                if brackground:
                    tasks_repository = kwargs.get('tasks_repository')
                    return make_background_task(
                        controller,
                        path_pattern,
                        tasks_repository=tasks_repository,
                    )

                else:
                    route = make_route(
                        path_pattern, MethodType[method], controller
                    )
                    return route.controller

            return wrapper

        return decorator


route = _RouteDecorator()



from typing import TypedDict

from jsondaora import jsondaora, typed_dict_asjson

from apidaora.asgi.app import asgi_app
from apidaora.asgi.responses import JSON_RESPONSE
from apidaora.asgi.router import Route, make_router
from apidaora.method import MethodType


@jsondaora
class You(TypedDict):
    name: str
    location: str


async def hello_controller(path_args, query_dict, headers, body):
    you = You(name=path_args['name'], location=query_dict['location'][0])
    body = typed_dict_asjson(you, You)
    return JSON_RESPONSE, body


route = Route(
    path_pattern='/hello/{name}',
    method=MethodType.GET,
    controller=hello_controller,
    has_path_args=True,
    has_query=True,
)
router = make_router([route])
app = asgi_app(router)
